/**************************************************************************
   mashup - arrange images and text in a layout and print

   Copyright 2008 2009 2010 2011 2012 Michael Cornelison
   source URL:  kornelix.squarespace.com
   contact: kornelix2@gmail.com
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "zfuncs.h"

#define fversion "mashup v.3.5"                                            //  version
#define flicense "Free software - GNU General Public License v.3"
#define fsource "http://kornelix.squarespace.com/mashup"

#define bilinear GDK_INTERP_BILINEAR                                       //  GDK shortcuts
#define nodither GDK_RGB_DITHER_NONE,0,0
#define colorspace GDK_COLORSPACE_RGB
#define pixbuf_scale gdk_pixbuf_scale_simple

#define maxtext 9999                                                       //  max. text cc
#define layoutsize 1600                                                    //  max. layout dimension
#define pack_tolerance 0.02                                                //  alignment tolerance for auto-packing

namespace print_image {                                                    //  page setup information from
   extern double  width, height, margins[4];                               //    print_image_page_setup()
   extern int     landscape;
}
double      pwidth, pheight;                                               //  paper dimensions in cm

GtkWidget   *mainWin, *drawWin, *mVbox;                                    //  main and drawing window
GdkPixbuf   *pxb_layout = 0;                                               //  layout pixbuf
GdkPixbuf   *pxb_drawwin = 0;                                              //  drawing window pixbuf
int         wwD = 1000, hhD = 700;                                         //  main window initial size
typedef     uchar * pixel;                                                 //  3 RGB values, 0-255 each
int         rectx = -1, recty, rectww, recthh;                             //  rectangle drawn over layout

zdialog     *zdtext = 0;                                                   //  active text dialog
zdialog     *zdrotate = 0;                                                 //  active rotate dialog
zdialog     *zdframe = 0;                                                  //  active frame dialog
zdialog     *zdtrimI = 0;                                                  //  active trim image dialog     v.3.3

char        imagefile[maxfcc] = "";                                        //  current image file
char        imagedirk[maxfcc] = "";                                        //  current image file directory

//  page layout data and default settings

int         layout_ww, layout_hh;                                          //  layout pixel dimensions
int         layout_pix = 400;                                              //  default image size in layout
int         layout_select = -1;                                            //  last selected image/text index
double      Rscale = 0.5;                                                  //  layout to drawing window
char        pendfile[maxfcc] = "";                                         //  pending image, add to layout
char        pendtext[maxtext+1] = "";                                      //  pending text, add to layout
char        textfont[60] = "sans 40";                                      //  layout text default font
int         textbw = 'b';                                                  //  default text: black on white
int         transp = 0;                                                    //  text background is transparent
int         frww = 0;                                                      //  default frame thickness   v.2.2
int         frRGB[3] = { 0, 0, 150 };                                      //  default frame color       v.2.2

struct      playout {                                                      //  image and text members of layout
   int         type;                                                       //  1: image, 2: text image
   int         textbw;                                                     //  text color, 'b' or 'w'
   int         transp;                                                     //  background is transparent
   char        *filespec;                                                  //  image filespec
   char        *textspec;                                                  //  text image text content
   GdkPixbuf   *pixbuf1;                                                   //  image pixbuf, full size
   int         fww, fhh;                                                   //  full image dimensions
   int         trmx, trmy, trww, trhh;                                     //  full image trim rectangle    v.3.3
   GdkPixbuf   *pixbuf2;                                                   //  layout pixbuf, rscale size
   int         px, py, pww, phh;                                           //  position and size in layout
   double      rscale;                                                     //  scale factor, image to layout
   double      angle;                                                      //  rotation, degrees CW      v.2.0
   int         frRGB[3];                                                   //  frame color               v.2.0
   int         frww;                                                       //  frame thickness           v.2.0
};
playout     layout[200];                                                   //  layout for all images + text
int         maxNF = 200;                                                   //  max. images + text in layout
int         NPL = 0;                                                       //  current layout count

//  GTK functions
int   gtkinitfunc(void *data);                                             //  GTK initz. function
void  mwpaint(GtkWidget *, cairo_t *);                                     //  window repaint - expose event
void  destroy();                                                           //  main window destroy signal
void  mouse_event(GtkWidget *, GdkEventButton *, void *);                  //  mouse event function
int   KBpress(GtkWidget *, GdkEventKey *, void *);                         //  KB key press event
int   KBrelease(GtkWidget *, GdkEventKey *, void *);                       //  KB key release event
void  drag_drop(int x, int y, char *text);                                 //  text drag/drop event   v.2.0.2
void  menufunc(GtkWidget *, const char *menu);                             //  menu function, main window

//  toolbar button functions
void  m_image();                                                           //  add images to layout
void  m_text();                                                            //  add text to layout
void  m_trimI();                                                           //  trim image
void  m_rotate();                                                          //  rotate image
void  m_frame();                                                           //  edit image frame
void  m_pack();                                                            //  pack images, close gaps
void  m_trimL();                                                           //  trim layout to fit images
void  m_save();                                                            //  save layout to file
void  m_format();                                                          //  select paper format
void  m_print();                                                           //  print finished layout 
void  m_help();                                                            //  show user guide
void  m_quit();                                                            //  quit

void  init_layout();                                                       //  page layout from paper size
void  set_pendfile(char *file);                                            //  set pending file to be added
void  add_layout_image(int mpx, int mpy);                                  //  add pending image to layout
void  add_layout_text(int mpx, int mpy);                                   //  add pending text to layout
void  save_imagedirk();                                                    //  save for next startup
void  load_imagedirk();                                                    //  load saved image directory

GdkPixbuf * make_layout(int final);                                        //  make layout pixbuf
GdkPixbuf * load_pixbuf(const char *file);                                 //  load pixbuf from file, remove alpha
GdkPixbuf * pixbuf_text(GtkWidget *, cchar *text, cchar *font, int bw);    //  create pixbuf containing text
void pixbuf_copy_alpha(GdkPixbuf *pxb1, int x1, int y1, int ww, int hh,    //  copy pixbuf with transparent color 
                       GdkPixbuf *pxb2, int x2, int y2, int rgb, int ff);
GdkPixbuf * pixbuf_add_frame(GdkPixbuf *pixbuf, int *rgb, int fww);        //  add frame around a pixbuf


/**************************************************************************/

//  main program

int main(int argc, char *argv[])
{
   GtkWidget   *mtbar;
   char        lang[8] = "";

   printf(fversion "\n");                                                  //  print version
   if (argc > 1 && strEqu(argv[1],"-v")) return 0;

   gtk_init(&argc,&argv);                                                  //  GTK command line options

   zinitapp("mashup",null);                                                //  get app directories

   for (int ii = 1; ii < argc; ii++)                                       //  command line options
   {
      if (strEqu(argv[ii],"-l") && argc > ii+1)                            //  -l language code
            strncpy0(lang,argv[++ii],7);
      else if (strEqu(argv[ii],"-i") && argc > ii+1)                       //  -i imageDirectory
            strcpy(imagedirk,argv[++ii]);
   }
   
   ZTXinit(lang);                                                          //  setup translations

   mainWin = gtk_window_new(GTK_WINDOW_TOPLEVEL);                          //  create main window
   gtk_window_set_title(GTK_WINDOW(mainWin),fversion);
   gtk_window_set_position(GTK_WINDOW(mainWin),GTK_WIN_POS_CENTER);
   gtk_window_set_default_size(GTK_WINDOW(mainWin),wwD,hhD);

   mVbox = gtk_vbox_new(0,0);                                              //  add vert. packing box
   gtk_container_add(GTK_CONTAINER(mainWin),mVbox);

   mtbar = create_toolbar(mVbox);                                          //  add toolbar and buttons  
   add_toolbar_button(mtbar,ZTX("+image"),ZTX("add image to layout"),"image.png",menufunc);
   add_toolbar_button(mtbar,ZTX("+text"),ZTX("add text to layout"),"text.png",menufunc);
   add_toolbar_button(mtbar,ZTX("trimI"),ZTX("trim image"),"trim.png",menufunc);
   add_toolbar_button(mtbar,ZTX("rotate"),ZTX("rotate image or text"),"rotate.png",menufunc);
   add_toolbar_button(mtbar,ZTX("frame"),ZTX("frame image"),"frame.png",menufunc);
   add_toolbar_button(mtbar,ZTX("pack"),ZTX("pack images"),"pack.png",menufunc);
   add_toolbar_button(mtbar,ZTX("trimL"),ZTX("trim layout"),"trim.png",menufunc);
   add_toolbar_button(mtbar,ZTX("save"),ZTX("save layout to file"),"gtk-save",menufunc);
   add_toolbar_button(mtbar,ZTX("format"),ZTX("layout and print format"),"format.png",menufunc);
   add_toolbar_button(mtbar,ZTX("print"),ZTX("print layout"),"print.png",menufunc);
   add_toolbar_button(mtbar,ZTX("quit"),ZTX("quit mashup"),"gtk-quit",menufunc);
   add_toolbar_button(mtbar,ZTX("help"),ZTX("help"),"gtk-help",menufunc);

   drawWin = gtk_drawing_area_new();                                       //  drawing window
   gtk_container_add(GTK_CONTAINER(mVbox),drawWin);                        //  add to main window

   G_SIGNAL(mainWin,"destroy",destroy,0);                                  //  connect events to main window
   G_SIGNAL(drawWin,"draw",mwpaint,0);                                     //  gtk3

   G_SIGNAL(mainWin,"key-press-event",KBpress,0);                          //  connect KB events
   G_SIGNAL(mainWin,"key-release-event",KBrelease,0);

   gtk_widget_add_events(drawWin,GDK_BUTTON_PRESS_MASK);                   //  connect mouse events
   gtk_widget_add_events(drawWin,GDK_BUTTON_RELEASE_MASK);
   gtk_widget_add_events(drawWin,GDK_BUTTON_MOTION_MASK);
   gtk_widget_add_events(drawWin,GDK_POINTER_MOTION_MASK);
   G_SIGNAL(drawWin,"button-press-event",mouse_event,0);
   G_SIGNAL(drawWin,"button-release-event",mouse_event,0);
   G_SIGNAL(drawWin,"motion-notify-event",mouse_event,0);

   drag_drop_connect(drawWin,drag_drop);                                   //  connect drag-drop event  v.2.0.2
   gtk_widget_show_all(mainWin);                                           //  show all widgets
   g_timeout_add(0,gtkinitfunc,0);                                         //  initz. function, run first

   gtk_main();                                                             //  process window events
   return 0;
}


//  initial function called from gtk_main() at startup

int gtkinitfunc(void *)
{
   FILE        *fid;
   char        file[200];
   int         nns = 0, nnf = 0;
   float       width, height;
   int         frame[4];
   
   snprintf(file,200,"%s/printformat",get_zuserdir());                     //  get saved format data  v.3.2
   fid = fopen(file,"r");
   if (fid) {
      nns = fscanf(fid,"print size: %f %f ",&width,&height);
      nnf = fscanf(fid,"frame: %d %d %d %d ",frame,frame+1,frame+2,frame+3);
      fclose(fid);
   }
   
   if (nns == 2) {                                                         //  get print width, height
      pwidth = width;
      pheight = height;
   }
   else {
      pwidth = 29.7;                                                       //  default A4 landscape
      pheight = 21.0;
   }
   
   if (nnf == 4) {                                                         //  get frame thickness and RGB color
      frww = frame[0];
      frRGB[0] = frame[1];
      frRGB[1] = frame[2];
      frRGB[2] = frame[3];
   }
   else {
      frww = 3;
      frRGB[0] = frRGB[1] = frRGB[2] = 255;
   }
   
   load_imagedirk();                                                       //  get prior image directory
   init_layout();                                                          //  set up page layout pixbuf
   return 0;
}


//  paint window when created, exposed, resized

void mwpaint(GtkWidget *, cairo_t *cr)
{
   GdkPixbuf   *pxbM;
   int         ww, hh;
   double      scalew, scaleh;
   
   if (pxb_layout) g_object_unref(pxb_layout);                             //  (re)make editing layout
   pxb_layout = make_layout(0);
   if (! pxb_layout) return;                                               //  v.3.1

   wwD = gtk_widget_get_allocated_width(drawWin);                          //  current drawing window size
   hhD = gtk_widget_get_allocated_height(drawWin);

   scalew = 1.0 * wwD / layout_ww;                                         //  scale layout to window
   scaleh = 1.0 * hhD / layout_hh;
   if (scalew < scaleh) Rscale = scalew;
   else Rscale = scaleh;
   ww = int(Rscale * layout_ww + 0.5);
   hh = int(Rscale * layout_hh + 0.5);
   pxbM = pixbuf_scale(pxb_layout,ww,hh,bilinear);
   if (! pxbM) zappcrash("cannot create pixbuf");

   gdk_cairo_set_source_pixbuf(cr,pxbM,0,0);
   cairo_paint(cr);
   g_object_unref(pxbM);
   
   if (rectx > -1) {
      cairo_set_line_width(cr,1);
      cairo_set_source_rgb(cr,0,0,0);
      cairo_rectangle(cr,rectx,recty,rectww,recthh);
      cairo_stroke(cr);
      cairo_set_source_rgb(cr,1,1,1);
      cairo_rectangle(cr,rectx+1,recty+1,rectww-2,recthh-2);
      cairo_stroke(cr);
   }

   return;
}


//  main window destroy signal

void destroy()
{
   printf("main window destroyed \n");
   m_quit();
   return;
}


//  mouse event function - capture mouse buttons and drag movements

void mouse_event(GtkWidget *, GdkEventButton *event, void *)
{
   static int     bdtime = 0, butime = 0;
   static int     Lmouse = 0, Rmouse = 0;
   static int     Fclick = 0, Fselect = 0, Fcorner = 0;
   static int     elapsed, mpx0 = 0, mpy0 = 0;
   static int     ii, kk, px, py, pww, phh;
   int            mpx, mpy, dx = 0, dy = 0;
   int            trww, trhh;
   int            x1, y1, x2, y2;
   double         rscale;
   playout        templayout;
   
   void  trimI_mouse_event(GtkWidget *, GdkEventButton *event, void *);

   if (zdtrimI) {
      trimI_mouse_event(null,event,null);                                  //  trim image underway    v.3.3
      return;
   }

   mpx = int(event->x);                                                    //  mouse position in window
   mpy = int(event->y);
   mpx = int(mpx / Rscale + 0.5);                                          //  in layout
   mpy = int(mpy / Rscale + 0.5);

   if (event->type == GDK_BUTTON_PRESS)
   {
      Lmouse = Rmouse = 0;
      if (event->button == 1) Lmouse++;                                    //  left or right mouse button
      if (event->button == 3) Rmouse++;
      bdtime = event->time;
      Fselect = Fcorner = 0;
      layout_select = -1;                                                  //  set no current item    v.2.0

      for (ii = NPL-1; ii >= 0; ii--)                                      //  search from top image down
      {
         px = layout[ii].px;                                               //  image position in layout
         py = layout[ii].py;
         pww = layout[ii].pww;                                             //  cursor inside an image?  v.2.0
         phh = layout[ii].phh;
         rscale = layout[ii].rscale;
         if (mpx < px) continue;
         if (mpx > px + pww) continue;
         if (mpy < py) continue;
         if (mpy > py + phh) continue;
         break;
      }
      if (ii >= 0) {                                                       //  yes, iith image/text selected
         layout_select = ii;                                               //  remember last selected object
         if (layout[ii].type == 1)                                         //  set current image file
            strcpy(imagefile,layout[ii].filespec);
         else strcpy(pendtext,layout[ii].textspec);                        //  or current text string
         Fselect = 1;
      }
      
      if (Fselect) 
      {
         mpx0 = mpx;                                                       //  set new drag origin
         mpy0 = mpy;
         
         if (layout[ii].type == 1) {                                             //  mouse in image corner?
            if (mpx < px + 0.2 * pww && mpy < py + 0.2 * phh) Fcorner = 1;       //  upper left
            if (mpx > px + 0.8 * pww && mpy < py + 0.2 * phh) Fcorner = 2;       //  upper right
            if (mpx > px + 0.8 * pww && mpy > py + 0.8 * phh) Fcorner = 3;       //  lower right
            if (mpx < px + 0.2 * pww && mpy > py + 0.8 * phh) Fcorner = 4;       //  lower left
         }

         if (layout[ii].type == 2) {                                             //  mouse in text corner?
            if (pww > phh) {                                                     //  (broader range for narrow edge)
               if (mpx < px + 0.2 * pww && mpy < py + 0.5 * phh) Fcorner = 1;    //  upper left
               if (mpx > px + 0.8 * pww && mpy < py + 0.5 * phh) Fcorner = 2;    //  upper right
               if (mpx > px + 0.8 * pww && mpy > py + 0.5 * phh) Fcorner = 3;    //  lower right
               if (mpx < px + 0.2 * pww && mpy > py + 0.5 * phh) Fcorner = 4;    //  lower left
            }

            else  {
               if (mpx < px + 0.5 * pww && mpy < py + 0.2 * phh) Fcorner = 1;    //  upper left
               if (mpx > px + 0.5 * pww && mpy < py + 0.2 * phh) Fcorner = 2;    //  upper right
               if (mpx > px + 0.5 * pww && mpy > py + 0.8 * phh) Fcorner = 3;    //  lower right
               if (mpx < px + 0.5 * pww && mpy > py + 0.8 * phh) Fcorner = 4;    //  lower left
            }
         }
      }

      return;
   }
   
   if (event->type == GDK_MOTION_NOTIFY)                                   //  capture mouse movement
   {
      if (! Lmouse) return;                                                //  only if left-mouse
      if (! Fselect) return;                                               //  v.3.3

      dx = mpx - mpx0;                                                     //  capture drag motion
      dy = mpy - mpy0;
      mpx0 = mpx;                                                          //  set new drag origin
      mpy0 = mpy;
      
      if (! Fcorner) {                                                     //  image move: add mouse drag
         layout[ii].px += dx;                                              //    to image position
         layout[ii].py += dy;
      }
      
      if (Fcorner)                                                         //  image resize: add mouse drag
      {                                                                    //    to image size
         rscale = layout[ii].rscale;
         trww = layout[ii].trww;
         trhh = layout[ii].trhh;

         x1 = layout[ii].px;                                               //  extent of layout image
         y1 = layout[ii].py;
         x2 = x1 + int(rscale * trww + 0.5);
         y2 = y1 + int(rscale * trhh + 0.5);

         if (Fcorner == 1) {                                               //  drag upper left corner
            if (trww > trhh) rscale = (rscale * trww - dx) / trww;
            else  rscale = (rscale * trhh - dy) / trhh;
            x1 = x2 - int(rscale * trww + 0.5);                            //  leave opposite corner fixed
            y1 = y2 - int(rscale * trhh + 0.5);
         }

         if (Fcorner == 2) {                                               //  upper right
            if (trww > trhh) rscale = (rscale * trww + dx) / trww;
            else  rscale = (rscale * trhh - dy) / trhh;
            x2 = x1 + int(rscale * trww + 0.5);
            y1 = y2 - int(rscale * trhh + 0.5);
         }

         if (Fcorner == 3) {                                               //  lower right
            if (trww > trhh) rscale = (rscale * trww + dx) / trww;
            else  rscale = (rscale * trhh + dy) / trhh;
            x2 = x1 + int(rscale * trww + 0.5);
            y2 = y1 + int(rscale * trhh + 0.5);
         }

         if (Fcorner == 4) {                                               //  lower left
            if (trww > trhh) rscale = (rscale * trww - dx) / trww;
            else  rscale = (rscale * trhh + dy) / trhh;
            x1 = x2 - int(rscale * trww + 0.5);
            y2 = y1 + int(rscale * trhh + 0.5);
         }
         
         if (rscale < 0.05 || rscale > 4.0) return;                        //  limit extremes
         
         if (trww > trhh) rscale = 1.0 * (x2 - x1) / trww;                 //  make precise
         else  rscale = 1.0 * (y2 - y1) / trhh;

         if (layout[ii].pixbuf2)                                           //  need new layout pixbuf
            g_object_unref(layout[ii].pixbuf2);
         layout[ii].pixbuf2 = 0;

         layout[ii].rscale = rscale;                                       //  will recreate from new data
         layout[ii].px = x1;
         layout[ii].py = y1;
      }
   }
   
   if (event->type == GDK_BUTTON_RELEASE)
   {
      Fclick = 0;
      butime = event->time;
      elapsed = butime - bdtime;                                           //  button down time, milliseconds
      if (elapsed < 500) Fclick = 1;                                       //  mouse clicked, no drag
      
      if (*pendtext && zdtext && Fclick && Lmouse)                         //  pending text and left-click  v.3.3
         add_layout_text(mpx,mpy);                                         //  add pending text image to layout

      else if (Fselect && Fclick && Lmouse) {                              //  image was left-clicked
         templayout = layout[ii];
         for (kk = ii; kk < NPL-1; kk++)                                   //  make topmost (will occlude others)
         {
            if (layout[kk+1].type > layout[ii].type) break;                //  all text above all images    v.1.6
            layout[kk] = layout[kk+1];
         }
         layout[kk] = templayout;
         ii = layout_select = kk;                                          //  new current image index      v.1.7
      }

      else if (Fselect && Fclick && Rmouse) {                              //  image was right-clicked
         if (layout[ii].filespec) zfree(layout[ii].filespec);
         if (layout[ii].textspec) zfree(layout[ii].textspec);
         if (layout[ii].pixbuf1) g_object_unref(layout[ii].pixbuf1);       //  delete image from layout
         if (layout[ii].pixbuf2) g_object_unref(layout[ii].pixbuf2);
         for (kk = ii; kk < NPL-1; kk++)                                   //  cover the hole
            layout[kk] = layout[kk+1];
         NPL--;
         layout_select = -1;                                               //  no selection active    v.2.0
      }

      else if (! Fselect && ! Fcorner && Fclick) {                         //  empty space was clicked
         if (*pendfile) add_layout_image(mpx,mpy);                         //  add pending file image to layout
         if (*pendtext) add_layout_text(mpx,mpy);                          //  add pending text image to layout
      }

      if (layout_select >= 0) {                                            //  if item selected and dialog 
         if (zdtext) zdialog_send_event(zdtext,"select");                  //    active, send select event  
         if (zdrotate) zdialog_send_event(zdrotate,"select");              //      to dialog     v.2.0
         if (zdframe) zdialog_send_event(zdframe,"select");
      }

      Fselect = Fcorner = Fclick = 0;                                      //  stop drag/resize
      Lmouse = Rmouse = 0;
   }

   gtk_widget_queue_draw(drawWin);                                         //  repaint window
   return;
}


//  prevent propagation of key-press events to toolbar buttons     v.1.7

int KBpress(GtkWidget *win, GdkEventKey *event, void *)
{
   return 1;
}


//  keyboard event - arrow keys >> 1-pixel movement of current image
//  GDK key symbols: /usr/include/gtk-2.0/gdk/gdkkeysyms.h

int KBrelease(GtkWidget *win, GdkEventKey *event, void *)                  //  process KB arrow keys   v.1.7
{
   int         ii, KBkey;

   if (layout_select < 0) return 1;
   ii = layout_select;

   KBkey = event->keyval;
   if (KBkey == GDK_KEY_Left) --layout[ii].px;                             //  move image in 1-pixel steps
   if (KBkey == GDK_KEY_Right) ++layout[ii].px;
   if (KBkey == GDK_KEY_Up) --layout[ii].py;
   if (KBkey == GDK_KEY_Down) ++layout[ii].py;
   
   gtk_widget_queue_draw(drawWin);
   return 1;
}


//  text drag and drop event - get filespec or text to add to layout       v.2.0.2

void drag_drop(int mpx, int mpy, char *text)
{
   mpx = int(mpx / Rscale + 0.5);                                          //  position in layout
   mpy = int(mpy / Rscale + 0.5);
   
   *pendfile = *pendtext = 0;                                              //  cancel prior pending items
   
   if (*text == '/')                                                       //  text is a filespec
   {
      set_pendfile(text);                                                  //  check if valid image file
      if (*pendfile) add_layout_image(mpx,mpy);                            //  add to layout at position
   }
   
   else                                                                    //  text is just text
   {      
      strncpy0(pendtext,text,maxtext);                                     //  save for layout insertion
      add_layout_text(mpx,mpy);
   }

   zfree(text);
   gtk_widget_queue_draw(drawWin);
   return;
}


//  process main window toolbar events

void menufunc(GtkWidget *, const char *menu)
{
   if (strEqu(menu,ZTX("+image"))) { m_image(); return; }
   if (strEqu(menu,ZTX("+text"))) { m_text(); return; }
   if (strEqu(menu,ZTX("trimI"))) { m_trimI(); return; }
   if (strEqu(menu,ZTX("rotate"))) { m_rotate(); return; }
   if (strEqu(menu,ZTX("frame"))) { m_frame(); return; }
   if (strEqu(menu,ZTX("pack"))) { m_pack(); return; }
   if (strEqu(menu,ZTX("trimL"))) { m_trimL(); return; }
   if (strEqu(menu,ZTX("save"))) { m_save(); return; }
   if (strEqu(menu,ZTX("format"))) { m_format(); return; }
   if (strEqu(menu,ZTX("print"))) { m_print(); return; }
   if (strEqu(menu,ZTX("quit"))) { m_quit(); return; }
   if (strEqu(menu,ZTX("help"))) { m_help(); return; }
   return;
}


//  choose image files and add to layout
//  file-chooser dialog box and response function
//  chosen file is saved for drawing window deposit via mouse-click

void m_image()
{
   GtkWidget      *vbox;

   int image_compl(GtkDialog *dwin, int arg, void *data);
   void image_preview(GtkFileChooser *, GtkWidget *pvwidget);

   GtkWidget   *dialog, *file_widget;
   GtkWidget   *pvwidget = gtk_image_new();

   dialog = gtk_dialog_new_with_buttons(ZTX("choose files"),GTK_WINDOW(mainWin),
                                         (GtkDialogFlags) 0,ZTX("done"),2,null);
   gtk_window_set_default_size(GTK_WINDOW(dialog),700,500);
   G_SIGNAL(dialog,"response",image_compl,0);

   vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));                 //  gtk3
   file_widget = gtk_file_chooser_widget_new(GTK_FILE_CHOOSER_ACTION_OPEN);
   gtk_box_pack_end(GTK_BOX(vbox),file_widget,1,1,0);
   gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(file_widget),imagedirk);
   gtk_file_chooser_set_show_hidden(GTK_FILE_CHOOSER(file_widget),0);

   gtk_file_chooser_set_preview_widget(GTK_FILE_CHOOSER(file_widget),pvwidget);
   G_SIGNAL(file_widget,"update-preview",image_preview,pvwidget);

   gtk_widget_show_all(dialog);
   return;
}

int image_compl(GtkDialog *dwin, int arg, void *data)                      //  done or cancel
{
   gtk_widget_destroy(GTK_WIDGET(dwin));
   return 0;
}

void image_preview(GtkFileChooser *file_widget, GtkWidget *pvwidget)       //  get file preview thumbnail
{
   GdkPixbuf   *thumbnail;
   char        *filename;
   
   filename = gtk_file_chooser_get_preview_filename(file_widget);
   if (! filename) return;

   thumbnail = get_thumbnail(filename,192);                                //  get 192x192 thumbnail
   if (thumbnail) {                                                        //  (cached or created)
      gtk_image_set_from_pixbuf(GTK_IMAGE(pvwidget),thumbnail);
      gtk_file_chooser_set_preview_widget_active(file_widget,1);
      g_object_unref(thumbnail);
      set_pendfile(filename);                                              //  save for layout insertion   v.2.0
   }
   else gtk_file_chooser_set_preview_widget_active(file_widget,0);

   g_free(filename);
   return;
}


//  add text to layout
//  get text dialog box and response function
//  entered text is saved for drawing window deposit via mouse-click

void m_text()
{
   int text_dialog_event(zdialog *, const char *event);

   int         ii;
   const char  *helptext = ZTX(" enter text, select font, \n"
                               " click empty space on layout ");
   if (zdtext) return;

   zdtext = zdialog_new(ZTX("enter text"),mainWin,ZTX("cancel"),null);
   zdialog_add_widget(zdtext,"label","lab1","dialog",helptext,"space=10");
   zdialog_add_widget(zdtext,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zdtext,"label","lab2","hb1",ZTX("text"),"space=3");        //  text [_______________]
   zdialog_add_widget(zdtext,"frame","fr1","hb1",0,"expand");
   zdialog_add_widget(zdtext,"edit","text","fr1",0,"space=5");
   zdialog_add_widget(zdtext,"hbox","hb2","dialog");                             //  (o) black   (o) white
   zdialog_add_widget(zdtext,"radio","black","hb2",ZTX("black"),"space=5");
   zdialog_add_widget(zdtext,"radio","white","hb2",ZTX("white"),"space=10");
   zdialog_add_widget(zdtext,"hbox","hb3","dialog");                             //  transparent [_]
   zdialog_add_widget(zdtext,"label","lab3","hb3",ZTX("transparent"),"space=8");
   zdialog_add_widget(zdtext,"check","transp","hb3",0);
   zdialog_add_widget(zdtext,"hbox","hb4","dialog",0,"space=10");                //  [erase text]  [select font] 
   zdialog_add_widget(zdtext,"button","erase","hb4",ZTX("erase text"));
   zdialog_add_widget(zdtext,"button","font","hb4",ZTX("select font"));

   ii = layout_select;                                                     //  last selected text image
   if (ii >= 0 && layout[ii].type == 2) {
      textbw = layout[ii].textbw;
      transp = layout[ii].transp;
      strcpy(pendtext,layout[ii].textspec);
      *pendfile = 0;
   }

   if (textbw == 'b') zdialog_stuff(zdtext,"black",1);                     //  use prior text color and
   else  zdialog_stuff(zdtext,"white",1);                                  //    transparency options
   if (transp) zdialog_stuff(zdtext,"transp",1);
   zdialog_stuff(zdtext,"text",pendtext);                                  //  prior or pending text

   zdialog_run(zdtext,text_dialog_event);                                  //  run dialog, parallel
   return;
}


//  dialog event function

int text_dialog_event(zdialog *zd,const char *event)
{
   GtkWidget   *font_dialog;
   char        *pp, text[1000] = "";
   int         ii, acolor;
   
   if (zd->zstat)                                                          //  cancel
   {
      zdialog_free(zdtext);
      return 0;
   }
   
   if (strEqu(event,"erase")) {                                            //  erase text entry box
      zdialog_stuff(zd,"text","");
      *pendtext = 0;
      return 0;
   }
   
   if (strEqu(event,"font")) {                                             //  select a new font
      font_dialog = gtk_font_selection_dialog_new(ZTX("select font"));
      gtk_font_selection_dialog_set_font_name(GTK_FONT_SELECTION_DIALOG(font_dialog),textfont);
      gtk_dialog_run(GTK_DIALOG(font_dialog));
      pp = gtk_font_selection_dialog_get_font_name(GTK_FONT_SELECTION_DIALOG(font_dialog));
      if (pp) strncpy0(textfont,pp,59);
      gtk_widget_destroy(GTK_WIDGET(font_dialog));
   }
   
   if (strEqu(event,"select")) {
      ii = layout_select;                                                  //  last selected text image
      if (ii < 0 || layout[ii].type != 2) return 0;
      zdialog_stuff(zd,"text",layout[ii].textspec);
   }

   zdialog_fetch(zd,"text",text,999);                                      //  capture entered text
   zdialog_fetch(zd,"transp",transp);                                      //  and transparency choice
   zdialog_fetch(zd,"black",acolor);                                       //  and black/white choice
   
   if (acolor) textbw = 'b';                                               //  black or white
   else textbw = 'w';

   if (*text) {
      if (NPL == maxNF-1) {
         zmessageACK(mainWin,ZTX("exceed %d files"),maxNF);
         return 0;
      }
      strncpy0(pendtext,text,maxtext);                                     //  save for layout insertion
      *pendfile = 0;
   }

   return 0;
}


//  trim the last-selected image within the layout

int      trimdat[4];

void m_trimI()                                                             //  new v.3.3
{
   int   trimI_dialog_event(zdialog *zd, const char *event);

   const char  *trimmess = ZTX("drag mouse down and right across an image \n"
                               "to select the visible area rectangle");
   if (zdtrimI) return;
   
   zdtrimI = zdialog_new(ZTX("trim image"),mainWin,ZTX("trim"),ZTX("restore"),ZTX("done"),null);
   zdialog_add_widget(zdtrimI,"label","labtrim","dialog",trimmess,"space=10");

   layout_select = -1;                                                     //  no selected image

   zdialog_run(zdtrimI,trimI_dialog_event);                                //  run dialog
   return;
}


//  trimI dialog event function

int trimI_dialog_event(zdialog *zd, const char *event)
{
   int      ii, zstat;
   
   zstat = zd->zstat;
   if (! zstat) return 0;                                                  //  wait for completion
   
   if (zstat <= 2) zd->zstat = 0;                                          //  trim or restore, keep dialog

   else {                                                                  //  done or cancel, kill dialog
      zdialog_free(zdtrimI);
      rectx = -1;                                                          //  clear layout trim rectangle
      gtk_widget_queue_draw(drawWin);
      return 0;
   }

   ii = layout_select;                                                     //  selected image
   if (ii < 0) return 0;

   if (zstat == 1) {                                                       //  trim
      layout[ii].trmx = trimdat[0];                                        //  set image trim margins
      layout[ii].trmy = trimdat[1];
      layout[ii].trww = trimdat[2];
      layout[ii].trhh = trimdat[3];
   }

   if (zstat == 2) {
      layout[ii].trmx = 0;                                                 //  restore
      layout[ii].trmy = 0;                                                 //  set full image size
      layout[ii].trww = layout[ii].fww;
      layout[ii].trhh = layout[ii].fhh;
   }
   
   if (layout[ii].pixbuf2)                                                 //  need new layout pixbuf
      g_object_unref(layout[ii].pixbuf2);
   layout[ii].pixbuf2 = 0;

   rectx = -1;
   gtk_widget_queue_draw(drawWin);
   return 0;
}


//  trimI mouse function

void trimI_mouse_event(GtkWidget *, GdkEventButton *event, void *)
{
   static int     bdown = 0, bdtime;
   static int     trmx, trmy, trww, trhh;
   int            butime, elapsed;
   int            ii, fww, fhh;
   int            px, py, pww, phh;
   int            mpx, mpy, mpxL, mpyL;
   double         rscale;

   mpx = int(event->x);                                                    //  mouse position in window
   mpy = int(event->y);

   mpxL = int(mpx / Rscale + 0.5);                                         //  in layout
   mpyL = int(mpy / Rscale + 0.5);

   if (event->type == GDK_BUTTON_PRESS)                                    //  start of mouse drag
   {
      layout_select = -1;                                                  //  no image selected
      rectx = -1;                                                          //  no rectangle overlay

      if (event->button != 1) return;                                      //  ignore unless left mouse

      bdtime = event->time;                                                //  button down time

      for (ii = NPL-1; ii >= 0; ii--)                                      //  search from top image down
      {
         px = layout[ii].px;
         py = layout[ii].py;
         pww = layout[ii].pww;                                             //  cursor inside an image?  v.2.0
         phh = layout[ii].phh;
         if (mpxL < px) continue;
         if (mpxL > px + pww) continue;
         if (mpyL < py) continue;
         if (mpyL > py + phh) continue;
         break;
      }

      if (ii >= 0) {                                                       //  yes, image/text selected
         if (layout[ii].type != 1) return;                                 //  ignore if not image

         layout_select = ii;                                               //  selected image
         fww = layout[ii].fww;                                             //  full image size
         fhh = layout[ii].fhh;
         
         if (layout[ii].angle) {                                           //  image was rotated
            layout[ii].angle = 0;                                          //  remove rotation
            layout[ii].trmx = layout[ii].trmy = 0;                         //  reset image trim rectangle
            layout[ii].trww = fww;                                         //    to be the entire image
            layout[ii].trhh = fhh;
            if (layout[ii].pixbuf2)                                        //  need new layout pixbuf
               g_object_unref(layout[ii].pixbuf2);
            layout[ii].pixbuf2 = 0;
         }

         trmx = mpx;                                                       //  new trim origin = mouse position
         trmy = mpy;
         trww = trhh = 20;                                                 //  initial trim rectangle = minimum
         bdown = 1;
      }
   }

   if (event->type == GDK_MOTION_NOTIFY)                                   //  capture mouse movement
   {
      if (! bdown) return;
      if (layout_select < 0) return;                                       //  no image selected

      if (mpx - trmx > 20) trww = mpx - trmx;                              //  track selected rectangle
      if (mpy - trmy > 20) trhh = mpy - trmy;
      
      rectx = trmx;                                                        //  draw trim rectangle on layout
      recty = trmy;
      rectww = trww;
      recthh = trhh;
   }
   
   if (event->type == GDK_BUTTON_RELEASE)                                  //  end of mouse drag
   {
      bdown = 0;
      if (layout_select < 0) return;                                       //  no selected image

      butime = event->time;
      elapsed = butime - bdtime;                                           //  button down time, milliseconds
      if (elapsed < 500) {
         rectx = -1;                                                       //  ignore very short drag
         layout_select = -1;
         return;
      }

      trmx = int(trmx / Rscale + 0.5);                                     //  scale mouse to layout coordinates
      trmy = int(trmy / Rscale + 0.5);
      trww = int(trww / Rscale + 0.5);
      trhh = int(trhh / Rscale + 0.5);

      ii = layout_select;                                                  //  trimmed image position and size
      frww = layout[ii].frww;
      px = layout[ii].px + frww;
      py = layout[ii].py + frww;
      fww = layout[ii].fww;
      fhh = layout[ii].fhh;
      rscale = layout[ii].rscale;

      trmx = (trmx - px) / rscale + 0.5;                                   //  get selected rectangle
      trmy = (trmy - py) / rscale + 0.5;                                   //    scaled to full image size
      trww = trww / rscale + 0.5;
      trhh = trhh / rscale + 0.5;

      trmx = trmx + layout[ii].trmx;                                       //  add trim origin to prior
      trmy = trmy + layout[ii].trmy;

      if (trmx < 0 || trmy < 0 ) {
         layout_select = -1;                                               //  rectangle not within image
         rectx = -1;
         return;
      }
      
      if (trmx + trww > fww) trww = fww - trmx;                            //  limit to image edge
      if (trmy + trhh > fhh) trhh = fhh - trmy;
      
      trimdat[0] = trmx;                                                   //  set trim data for dialog
      trimdat[1] = trmy;
      trimdat[2] = trww;
      trimdat[3] = trhh;
   }
   
   gtk_widget_queue_draw(drawWin);                                         //  repaint window
   return;
}


//  rotate last-selected image or text through arbitrary angle             //  new  v.2.0

void m_rotate()
{
   int    rotate_dialog_event(zdialog *, const char *event);

   char        text[20];
   double      angle = 0;
   
   if (zdrotate) return;

   zdrotate = zdialog_new(ZTX("rotate image"),mainWin,ZTX("done"),ZTX("cancel"),null);
   zdialog_add_widget(zdrotate,"label","labdeg","dialog",ZTX("degrees"),"space=5");
   zdialog_add_widget(zdrotate,"hbox","hb1","dialog",0,"homog|space=5");
   zdialog_add_widget(zdrotate,"vbox","vb1","hb1",0,"space=5");
   zdialog_add_widget(zdrotate,"vbox","vb2","hb1",0,"space=5");
   zdialog_add_widget(zdrotate,"vbox","vb3","hb1",0,"space=5");
   zdialog_add_widget(zdrotate,"vbox","vb4","hb1",0,"space=5");
   zdialog_add_widget(zdrotate,"button"," +0.1  ","vb1"," + 0.1 ");        //  button name is increment to use
   zdialog_add_widget(zdrotate,"button"," -0.1  ","vb1"," - 0.1 ");
   zdialog_add_widget(zdrotate,"button"," +1.0  ","vb2"," + 1   ");
   zdialog_add_widget(zdrotate,"button"," -1.0  ","vb2"," - 1   ");
   zdialog_add_widget(zdrotate,"button"," +10.0 ","vb3"," + 10  ");
   zdialog_add_widget(zdrotate,"button"," -10.0 ","vb3"," - 10  ");
   zdialog_add_widget(zdrotate,"button"," +90.0 ","vb4"," + 90  ");
   zdialog_add_widget(zdrotate,"button"," -90.0 ","vb4"," - 90  ");
   zdialog_add_widget(zdrotate,"hbox","hb2","dialog",0,"space=10");

   if (layout_select >= 0) angle = layout[layout_select].angle;
   sprintf(text,ZTX("degrees: %.1f"),angle);                               //  update dialog angle display
   zdialog_stuff(zdrotate,"labdeg",text);

   zdialog_run(zdrotate,rotate_dialog_event);                              //  run dialog - parallel
   return;
}


//  dialog event function

int rotate_dialog_event(zdialog *zd, const char *event)
{
   int         ii, err;
   char        text[20];
   double      angle, incr;
   
   if (zd->zstat)                                                          //  done or cancel
   {
      if (zd->zstat != 1 && layout_select >= 0) 
         layout[layout_select].angle = 0;
      zdialog_free(zdrotate);                                              //  v.3.3
      gtk_widget_queue_draw(drawWin);
      return 0;
   }

   ii = layout_select;                                                     //  use last selected object
   if (ii < 0) return 0;

   if (strpbrk(event,"+-")) {
      err = convSD(event,incr);                                            //  button name is increment to use
      if (err) return 0;
   }
   else return 0;

   angle = layout[ii].angle + incr;

   sprintf(text,ZTX("degrees: %.1f"),angle);                               //  update dialog angle display
   zdialog_stuff(zd,"labdeg",text);
   
   if (angle >= 360) angle -= 360;
   if (angle <= -360) angle += 360;
   if (fabs(angle) < 0.01) angle = 0;
   layout[ii].angle = angle;

   if (layout[ii].pixbuf2)                                                 //  need new layout pixbuf
      g_object_unref(layout[ii].pixbuf2);
   layout[ii].pixbuf2 = 0;

   gtk_widget_queue_draw(drawWin);                                         //  update layout in window
   return 1;
}


//  add or change the frame around an image

void m_frame()
{
   int    frame_dialog_event(zdialog *, const char * event);

   int      ii;
   char     color[20];

   if (zdframe) return;                                                    //  already active

   ii = layout_select;                                                     //  use last selected image
   if (ii < 0) return;
   if (layout[ii].type != 1) return;
   
   frww = layout[ii].frww;                                                 //  set frame data from image
   memcpy(frRGB,layout[ii].frRGB,3*sizeof(int));

   zdframe = zdialog_new(ZTX("edit frame"),mainWin,ZTX("done"),ZTX("cancel"),null);
   zdialog_add_widget(zdframe,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdframe,"label","lab1","hb1",ZTX("thickness"));
   zdialog_add_widget(zdframe,"spin","thick","hb1","0|20|1|8","scc=2");
   zdialog_add_widget(zdframe,"label","space","hb1","","space=10");
   zdialog_add_widget(zdframe,"label","lab2","hb1",ZTX("color"));
   zdialog_add_widget(zdframe,"colorbutt","color","hb1","100|100|100");
   zdialog_add_widget(zdframe,"hbox","hb2","dialog");
   zdialog_add_widget(zdframe,"radio","this","hb2","this image");
   zdialog_add_widget(zdframe,"radio","all","hb2","all images","space=10");
   
   zdialog_stuff(zdframe,"thick",layout[ii].frww);                         //  stuff frame data into dialog
   snprintf(color,19,"%d|%d|%d",frRGB[0],frRGB[1],frRGB[2]);
   zdialog_stuff(zdframe,"color",color);

   zdialog_run(zdframe,frame_dialog_event);                                //  run dialog - parallel
   return;
}


//  dialog event function

int frame_dialog_event(zdialog *zd, const char * event)
{
   int         ii, Fall;
   char        color[20];
   const char  *pp;
   
   if (zd->zstat)
   {
      zdialog_free(zdframe);
      gtk_widget_queue_draw(drawWin);
      return 0;
   }

   ii = layout_select;                                                     //  use last selected image 
   if (ii < 0) return 0;
   if (layout[ii].type != 1) return 0;

   frww = layout[ii].frww;                                                 //  set frame data from image
   memcpy(frRGB,layout[ii].frRGB,3*sizeof(int));

   if (strEqu(event,"select")) {                                           //  stuff dialog from image frame
      zdialog_stuff(zd,"thick",frww);
      snprintf(color,19,"%d|%d|%d",frRGB[0],frRGB[1],frRGB[2]);
      zdialog_stuff(zd,"color",color);
      return 1;
   }

   zdialog_fetch(zd,"thick",frww);                                         //  get frame data from dialog
   zdialog_fetch(zd,"color",color,19);

   pp = strField(color,'|',1);
   if (pp) frRGB[0] = atoi(pp);
   pp = strField(color,'|',2);
   if (pp) frRGB[1] = atoi(pp);
   pp = strField(color,'|',3);
   if (pp) frRGB[2] = atoi(pp);
   
   zdialog_fetch(zd,"all",Fall);

   if (! Fall)                                                             //  update selected image
   {
      layout[ii].frww = frww;                                              //  set image frame data
      layout[ii].frRGB[0] = frRGB[0];
      layout[ii].frRGB[1] = frRGB[1];
      layout[ii].frRGB[2] = frRGB[2];
      if (layout[ii].pixbuf2)                                              //  need new layout pixbuf
         g_object_unref(layout[ii].pixbuf2);
      layout[ii].pixbuf2 = 0;
   }
   
   if (Fall)                                                               //  update all images      v.2.6
   {
      for (ii = 0; ii < NPL; ii++)
      {
         if (layout[ii].type != 1) continue;
         layout[ii].frww = frww;
         layout[ii].frRGB[0] = frRGB[0];
         layout[ii].frRGB[1] = frRGB[1];
         layout[ii].frRGB[2] = frRGB[2];
         if (layout[ii].pixbuf2)
            g_object_unref(layout[ii].pixbuf2);
         layout[ii].pixbuf2 = 0;
      }
   }

   gtk_widget_queue_draw(drawWin);
   return 1;
}


//  align and pack the images

void m_pack()                                                              //  overhauled v.3.3
{
   #define pnear(x,y) (abs((x)-(y)) < packtolr)                            //  close, within packing tolerance

   int         ii, jj, packtolr;
   int         iiL, iiR, iiT, iiB;                                         //  image position edges,
   int         jjL, jjR, jjT, jjB;                                         //    left, right, top, bottom
   int         frww, Fscale;
   double      rscale;
   
   packtolr = layout_ww;                                                   //  [pack] alignment tolerance,
   if (layout_hh > packtolr) packtolr = layout_hh;                         //    fraction of layout scale
   packtolr = packtolr * pack_tolerance;
   
   for (ii = 0; ii < NPL; ii++)
   {
      if (layout[ii].type == 2) continue;                                  //  pack to top and left layout edges
      if (pnear(layout[ii].px,0)) layout[ii].px = 0;
      if (pnear(layout[ii].py,0)) layout[ii].py = 0;
   }

   for (ii = 0; ii < NPL; ii++)                                            //  compare all image pairs
   for (jj = 0; jj < NPL; jj++)
   {
      if (layout[ii].type == 2) continue;
      if (layout[jj].type == 2) continue;
      if (jj == ii) continue;
      
      iiL = layout[ii].px;                                                 //  image ii edges
      iiR = iiL + layout[ii].pww;
      iiT = layout[ii].py;
      iiB = iiT + layout[ii].phh;
      
      jjL = layout[jj].px;                                                 //  image jj edges
      jjR = jjL + layout[jj].pww;
      jjT = layout[jj].py;
      jjB = jjT + layout[jj].phh;
      
      frww = layout[ii].frww + layout[jj].frww;                            //  mean ii/jj frame thickness
      frww = frww / 2;
      
      if (pnear(jjL,iiR) && jjT < iiB && jjB > iiT) {                      //  if jj left near ii right
         jjL = iiR - frww;                                                 //    move jj to butt ii
         jjR = jjL + layout[jj].pww;
      }

      if (pnear(jjT,iiB) && jjL < iiR && jjR > iiL) {                      //  if jj top near ii bottom
         jjT = iiB - frww;                                                 //    move jj to butt ii
         jjB = jjT + layout[jj].phh;
      }

      if (pnear(jjL,iiR) && pnear(jjT,iiT)) {                              //  if jj top left near ii top right
         jjL = iiR - frww;                                                 //    move jj corner to ii corner
         jjT = iiT;
         jjR = jjL + layout[jj].pww;
         jjB = jjT + layout[jj].phh;
      }
      
      if (pnear(jjL,iiL) && pnear(jjT,iiB)) {                              //  if jj top left near ii bottom left
         jjT = iiB - frww;                                                 //    move jj corner to ii corner
         jjL = iiL;
         jjR = jjL + layout[jj].pww;
         jjB = jjT + layout[jj].phh;
      }

      if (pnear(jjL,iiR) && pnear(jjB,iiB)) {                              //  jj bottom left near ii bottom right
         jjL = iiR - frww;                                                 //    move jj corner to ii corner
         jjB = iiB;
         jjR = jjL + layout[jj].pww;
      }
      
      if (pnear(jjT,iiB) && pnear(jjR,iiR)) {                              //  jj top right near ii bottom right
         jjT = iiB - frww;                                                 //    move jj corner to ii corner
         jjR = iiR;
         jjB = jjT + layout[jj].phh;
      }
      
      layout[jj].px = jjL;                                                 //  reposition image jj
      layout[jj].py = jjT;

      Fscale = 0;      

      if (jjR - jjL != layout[jj].pww) {                                   //  image jj width changed
         rscale = (jjR - jjL - 2 * layout[jj].frww);
         rscale = rscale / layout[jj].trww;
         Fscale++;
      }

      else if (jjB - jjT != layout[jj].phh) {                              //  image jj height changed
         rscale = (jjB - jjT - 2 * layout[jj].frww);
         rscale = rscale / layout[jj].trhh;
         Fscale++;
      }

      if (Fscale) {
         layout[jj].rscale = rscale;                                       //  rescale image jj
         if (layout[jj].pixbuf2) 
            g_object_unref(layout[jj].pixbuf2);
         layout[jj].pixbuf2 = 0;
      }

      if (Fscale) goto quit;
   }

   quit:  
   gtk_widget_queue_draw(drawWin);                                         //  refresh layout window
}


//  trim the layout to minimum size that fits the contained images

void m_trimL()                                                             //  new v.2.6
{
   int         ii, ww, hh, maxww, maxhh;

   if (! NPL) return;                                                      //  empty layout     bugfix v.3.1

   maxww = maxhh = 0;

   for (ii = 0; ii < NPL; ii++)
   {
      ww = layout[ii].px + layout[ii].pww;                                 //  bugfix  v.3.3
      if (ww > maxww) maxww = ww;
      hh = layout[ii].py + layout[ii].phh;
      if (hh > maxhh) maxhh = hh;
   }

   layout_ww = maxww;
   layout_hh = maxhh;
   gtk_widget_queue_draw(drawWin);
   return;
}


//  save layout to a file

void m_save()
{
   GdkPixbuf   *pixbufL;
   GError      *gerror = 0;
   char        *file1, file2[maxfcc], *pp;
   
   file1 = zgetfile1(ZTX("save file"),"save",imagedirk);                   //  query user for file
   if (! file1) return;
   
   strncpy0(file2,file1,maxfcc-6);
   zfree(file1);
   pp = strrchr(file2,'/');                                                //  force .png extension
   if (pp) pp = strrchr(pp,'.');
   if (! pp) pp = file2 + strlen(file2);
   strcpy(pp,".png");

   pixbufL = make_layout(1);                                               //  make 200% layout
   if (! pixbufL) return;                                                  //  v.3.1

   gdk_pixbuf_save(pixbufL,file2,"png",&gerror,null);
   g_object_unref(pixbufL);

   if (gerror) printf("gdk_pixbuf_save error: %s \n",gerror->message);     //  diagnose error    v.3.3
   return;
}


//  dialog for layout and print page format

void m_format()                                                            //  v.3.2
{
   print_image_paper_setup(mainWin);                                       //  select paper size and orientation
   
   pwidth = print_image::width;
   pheight = print_image::height;

   init_layout();                                                          //  set layout to match
   return;
}


//  dialog for print setup and print

void m_print()
{
   GdkPixbuf   *pixbufL;
   GError      *gerror = 0;                                                //  bugfix   v.2.3.1
   char        printfile[200];

   pixbufL = make_layout(1);                                               //  make 200% layout
   if (! pixbufL) return;                                                  //  v.3.1

   snprintf(printfile,200,"%s/printfile.png",get_zuserdir());              //  make .png print file   v.3.3
   gdk_pixbuf_save(pixbufL,printfile,"png",&gerror,null);
   g_object_unref(pixbufL);
   
   print_image_file(printfile);                                            //  print                           
   
   return;
}


//  display help file in separate process

void m_help()
{
   showz_userguide();
   return;
}


//  quit - exit program

void m_quit()
{
   char     file[200];
   FILE     *fid;

   save_imagedirk();                                                       //  save image directory

   snprintf(file,200,"%s/printformat",get_zuserdir());
   fid = fopen(file,"w");                                                  //  update paper format file
   if (fid) {
      fprintf(fid,"print size: %.1f %.1f \n",pwidth,pheight);
      fprintf(fid,"frame: %d %d %d %d \n",frww,frRGB[0],frRGB[1],frRGB[2]);
      fclose(fid);
   }

   gtk_main_quit();
   return;
}


//  set up page layout pixbuf based on layout dimensions

void init_layout()
{
   if (pwidth > pheight) {
      layout_ww = layoutsize;
      layout_hh = 1.0 * layout_ww * pheight / pwidth + 0.5;
   }
   else {
      layout_hh = layoutsize;
      layout_ww = 1.0 * layout_hh * pwidth / pheight + 0.5;
   }
   
   pxb_layout = gdk_pixbuf_new(colorspace,0,8,layout_ww,layout_hh);        //  pixbuf for page layout
   if (! pxb_layout) {
      zmessageACK(mainWin,"cannot create pixbuf %d %d",layout_ww,layout_hh);
      return;
   }

   wwD = int(Rscale * layout_ww);                                          //  matching drawing window size
   hhD = int(Rscale * layout_hh + 60);
   gtk_window_resize(GTK_WINDOW(mainWin),wwD,hhD);

   return;
}


//  verify chosen file and make pending add to layout
 
void set_pendfile(char *file)
{
   GdkPixbuf      *pixbuf;
   char           *pp;

   if (NPL == maxNF-1) {
      zmessageACK(mainWin,ZTX("exceed %d files"),maxNF);
      return;
   }
   
   pixbuf = load_pixbuf(file);                                             //  test file is legit. image file
   if (! pixbuf) {
      *pendfile = 0;                                                       //  simply ignore    v.2.0
      return;
   }
   g_object_unref(pixbuf);

   strcpy(pendfile,file);                                                  //  save filespec
   *pendtext = 0;

   strcpy(imagefile,file);                                                 //  set curr. image file
   strcpy(imagedirk,file);                                                 //  and image directory
   pp = (char *) strrchr(imagedirk,'/');
   if (pp) *pp = 0;
   
   gtk_window_present(GTK_WINDOW(mainWin));                                //  bring to foreground
   return;
}


//  add a new image file to the layout at designated position

void add_layout_image(int mpx, int mpy)
{
   GdkPixbuf   *pixbuf1;
   int         ii, kk;

   pixbuf1 = load_pixbuf(pendfile);
   if (! pixbuf1) zappcrash("cannot create pixbuf");
   for (ii = 0; ii < NPL; ii++)                                            //  insert at end of images
      if (layout[ii].type != 1) break;                                     //  (topmost file image)
   for (kk = NPL; kk > ii; kk--)
      layout[kk] = layout[kk-1];
   NPL++;
   layout_select = ii;
   layout[ii].type = 1;
   layout[ii].textbw = 0;
   layout[ii].transp = 0;
   layout[ii].filespec = strdupz(pendfile);
   layout[ii].textspec = 0;
   layout[ii].pixbuf1 = pixbuf1;
   layout[ii].pixbuf2 = 0;
   layout[ii].fww = gdk_pixbuf_get_width(pixbuf1);
   layout[ii].fhh = gdk_pixbuf_get_height(pixbuf1);
   layout[ii].trmx = layout[ii].trmy = 0;                                  //  no trim margins     v.3.3
   layout[ii].trww = layout[ii].fww;
   layout[ii].trhh = layout[ii].fhh;
   layout[ii].px = mpx;
   layout[ii].py = mpy;
   layout[ii].pww = layout[ii].phh = 0;
   layout[ii].rscale = 1.0 * layout_pix / layout[ii].fww;
   layout[ii].angle = 0;                                                   //  v.2.0
   memcpy(layout[ii].frRGB,frRGB,3*sizeof(int));                           //  v.2.0
   layout[ii].frww = frww;                                                 //  v.2.0
   *pendfile = 0;                                                          //  v.2.0
   return;
}


//  add a new text image to the layout at designated position

void add_layout_text(int mpx, int mpy)
{
   GdkPixbuf   *pixbuf1;
   int         ii;

   pixbuf1 = pixbuf_text(drawWin,pendtext,textfont,textbw);
   if (! pixbuf1) zappcrash("cannot create pixbuf");
   ii = NPL++;                                                             //  topmost text image
   layout_select = ii;
   layout[ii].type = 2;
   layout[ii].textbw = textbw;
   layout[ii].transp = transp;
   layout[ii].textspec = strdupz(pendtext);
   layout[ii].filespec = 0;
   layout[ii].pixbuf1 = pixbuf1;
   layout[ii].pixbuf2 = 0;
   layout[ii].fww = gdk_pixbuf_get_width(pixbuf1);
   layout[ii].fhh = gdk_pixbuf_get_height(pixbuf1);
   layout[ii].trmx = layout[ii].trmy = 0;                                  //  no trim margins     v.3.3
   layout[ii].trww = layout[ii].fww;
   layout[ii].trhh = layout[ii].fhh;
   layout[ii].px = mpx;
   layout[ii].py = mpy;
   layout[ii].pww = layout[ii].phh = 0;
   layout[ii].rscale = 0.5;
   layout[ii].angle = 0;                                                   //  v.2.0
   layout[ii].frww = 0;                                                    //  v.2.0
   *pendtext = 0;                                                          //  v.2.0
   return;
}


//  save current image directory upon exit, reload upon startup            //  v.1.2
//  directory is saved in file $HOME/.mashup/image_directory

void save_imagedirk()
{
   char     command[maxfcc+100];
   int      err;
   
   snprintf(command,maxfcc+99,"echo %s > %s/image_directory",              //  save imagedirk to file
                                    imagedirk, get_zuserdir());
   err = system(command);
   if (err) printf("error: %s \n",wstrerror(err));   

   return;
}

void load_imagedirk()
{
   int            err;
   FILE           *fid;
   struct stat    statdat;
   char           dirbuff[maxfcc], *pp;

   pp = getcwd(imagedirk,maxfcc-1);                                        //  default is current directory

   snprintf(dirbuff,maxfcc-1,"%s/image_directory",get_zuserdir());         //  read saved file
   fid = fopen(dirbuff,"r");
   if (! fid) return;
   pp = fgets_trim(dirbuff,maxfcc-1,fid,1);
   fclose(fid);
   if (! pp) return;
   err = stat(dirbuff,&statdat);                                           //  contains valid directory name?
   if (err) return;
   if (! S_ISDIR(statdat.st_mode)) return;
   strcpy(imagedirk,dirbuff);                                              //  yes, use it
   return;
}


//  make a layout pixbuf containing all scaled images
//  final = 0 = pixbuf for editing: 1x size and scaled image pixbufs are saved and reused
//  final = 1 = pixbuf to print or save: 2x size and scaled image pixbufs are discarded

GdkPixbuf * make_layout(int final)
{
   GdkPixbuf   *pixbufL, *pixbuf1, *pixbuf2, *pixbuf3;
   int         type, textbw, transp, ftext;
   int         ii, lscale, lww, lhh;
   int         fww, fhh, trmx, trmy, trww, trhh;
   int         px, py, pww, phh;
   int         orgx, orgy;
   int         rgb, rcolor, acolor;
   int         *frRGB;
   int         frww;
   double      angle, rscale;

   lscale = 1;
   if (final) lscale = 2;                                                  //  use 2x size for final layout

   lww = lscale * layout_ww;                                               //  layout size
   lhh = lscale * layout_hh;

   pixbufL = gdk_pixbuf_new(colorspace,0,8,lww,lhh);                       //  layout pixbuf
   if (! pixbufL) zappcrash("cannot create pixbuf");                       //  v.3.1
   gdk_pixbuf_fill(pixbufL,(unsigned) 0xffffffff);

   for (ii = 0; ii < NPL; ii++)                                            //  copy image pixbufs
   {                                                                       //    into layout pixbuf
      type = layout[ii].type;
      textbw = layout[ii].textbw;
      transp = layout[ii].transp;
      rscale = layout[ii].rscale;
      angle = layout[ii].angle;
      frRGB = layout[ii].frRGB;
      frww = lscale * layout[ii].frww;                                     //  bugfix  v.3.3
      pixbuf1 = layout[ii].pixbuf1;                                        //  image pixbuf
      fww = layout[ii].fww;                                                //  full size image dimensions
      fhh = layout[ii].fhh;
      trmx = layout[ii].trmx;                                              //  image trim rectangle   v.3.3
      trmy = layout[ii].trmy;
      trww = layout[ii].trww;
      trhh = layout[ii].trhh;
      pixbuf2 = layout[ii].pixbuf2;                                        //  scaled pixbuf in layout
      px = lscale * layout[ii].px;                                         //  image position in layout
      py = lscale * layout[ii].py;
      pww = lscale * int(rscale * trww + 0.5);                             //  image size in layout
      phh = lscale * int(rscale * trhh + 0.5);                             //  (without frame or rotation)

      if (type == 1) {                                                     //  image                  v.2.0
         if (angle == 0) rcolor = acolor = 0;                              //  no angle, no rotate border
         else rcolor = acolor = 'w';                                       //  border and alpha color = white
         ftext = 0;
      }
      else {                                                               //  text
         if (textbw == 'b') rcolor = acolor = 'w';                         //  rotate border and alpha color
         else  rcolor = acolor = 'b';                                      //    opposite of text color
         if (! transp) acolor = 0;                                         //  opaque background, no alpha color
         ftext = 1;
      }

      if (! pixbuf2 || final)
      {
         if (trww < fww || trhh < fhh) {                                   //  trim image               v.3.3
            pixbuf3 = gdk_pixbuf_new(colorspace,0,8,trww,trhh);
            if (! pixbuf3) zappcrash("cannot create pixbuf");
            gdk_pixbuf_copy_area(pixbuf1,trmx,trmy,trww,trhh,pixbuf3,0,0);
            pixbuf2 = pixbuf_scale(pixbuf3,pww,phh,bilinear);              //  scale to layout
            g_object_unref(pixbuf3);
         }
         else
            pixbuf2 = pixbuf_scale(pixbuf1,pww,phh,bilinear);              //  scale full image to layout
         
         if (frww) {                                                       //  v.3.3
            pixbuf3 = pixbuf_add_frame(pixbuf2,frRGB,frww);                //  add frame around image    v.2.0
            g_object_unref(pixbuf2);
            pixbuf2 = pixbuf3;
         }
         
         if (angle) {                                                      //  rotate if needed    v.2.0
            if (rcolor == 'b') rgb = 0;                                    //  (becomes larger than iww, ihh)
            else rgb = 255;
            pixbuf3 = gdk_pixbuf_rotate(pixbuf2,angle,rgb);
            g_object_unref(pixbuf2);
            pixbuf2 = pixbuf3;
         }
      }

      pww = gdk_pixbuf_get_width(pixbuf2);                                 //  image size with frame
      phh = gdk_pixbuf_get_height(pixbuf2);                                //    and poss. rotation
      layout[ii].pww = pww;
      layout[ii].phh = phh;

      orgx = orgy = 0;                                                     //  cut-off beyond layout edge
      if (px < 0) {
         orgx = -px;
         px = 0;
      }
      if (py < 0) {
         orgy = -py;
         py = 0;
      }

      if (px + pww > lww) pww = lww - px;
      if (py + phh > lhh) phh = lhh - py;

      if (acolor) {                                                        //  paste image into layout
         if (acolor == 'b') rgb = 0;                                       //    with transparency
         else rgb = 255;
         pixbuf_copy_alpha(pixbuf2,orgx,orgy,pww-orgx,phh-orgy,pixbufL,px,py,rgb,ftext);
      }
      else gdk_pixbuf_copy_area(pixbuf2,orgx,orgy,pww-orgx,phh-orgy,pixbufL,px,py);

      if (! final) layout[ii].pixbuf2 = pixbuf2;                           //  retain for efficiency
      else g_object_unref(pixbuf2);
   }
   
   return pixbufL;
}


//  validate an image file and load pixbuf from file                       //  v.1.2
//  if an alpha channel is present, remove it

GdkPixbuf * load_pixbuf(const char *file)
{
   GdkPixbuf   *pxb91 = 0, *pxb92 = 0;
   GError      *gerror = 0;                                                //  bugfix   v.2.3.1
   int         ww91, hh91, rs91, rs92;
   int         nch, nbits, px, py, alfa;
   pixel       ppix91, pix91, ppix92, pix92;

   pxb91 = gdk_pixbuf_new_from_file(file,&gerror);                         //  validate file and load pixbuf
   if (! pxb91) return 0;

   nch = gdk_pixbuf_get_n_channels(pxb91);
   nbits = gdk_pixbuf_get_bits_per_sample(pxb91);
   alfa = gdk_pixbuf_get_has_alpha(pxb91);

   if (nch < 3 || nbits != 8) {                                            //  must be 3 or 4 channels
      g_object_unref(pxb91);                                               //  and 8 bits per channel
      return 0;
   }
   
   if (! alfa) return pxb91;                                               //  no alpha channel, 3 channels

   ww91 = gdk_pixbuf_get_width(pxb91);                                     //  copy without alpha    v.1.2
   hh91 = gdk_pixbuf_get_height(pxb91);
   rs91 = gdk_pixbuf_get_rowstride(pxb91);
   ppix91 = gdk_pixbuf_get_pixels(pxb91);

   pxb92 = gdk_pixbuf_new(colorspace,0,8,ww91,hh91);
   if (! pxb92) zappcrash("cannot create pixbuf");
   rs92 = gdk_pixbuf_get_rowstride(pxb92);
   ppix92 = gdk_pixbuf_get_pixels(pxb92);

   for (py = 0; py < hh91; py++)
   for (px = 0; px < ww91; px++)
   {
      pix91 = ppix91 + rs91 * py + 4 * px;
      pix92 = ppix92 + rs92 * py + 3 * px;
      pix92[0] = pix91[0];
      pix92[1] = pix91[1];
      pix92[2] = pix91[2];
   }

   g_object_unref(pxb91);
   return pxb92;
}


//  create a pixbuf containing text
//  bw = 'b'/'w' for black/white text on white/black background

GdkPixbuf * pixbuf_text(GtkWidget *drawWin, cchar *text, cchar *font, int bw)
{
   PangoFontDescription *pfont;
   PangoLayout          *playout;
   cairo_surface_t      *surface;
   cairo_t              *cr;
   GdkPixbuf            *pixbuf;
   int                  ww, hh, px, py, rs;
   uint8                *cpix, *ppix, *pix1, *pix2;

   pfont = pango_font_description_from_string(font);
   playout = gtk_widget_create_pango_layout(drawWin,null);
   pango_layout_set_font_description(playout,pfont);
   pango_layout_set_text(playout,text,-1);
   pango_layout_get_pixel_size(playout,&ww,&hh);
   if (! ww) return 0;
   ww += 2;                                                                //  sometimes a little too small
   hh += 2;

   surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,ww,hh);         //  cairo output image        gtk3
   cr = cairo_create(surface);
   pango_cairo_show_layout(cr,playout);                                    //  write text layout to image
   cpix = cairo_image_surface_get_data(surface);                           //  output text image pixels
   
   pixbuf = gdk_pixbuf_new(colorspace,0,8,ww,hh);                          //  create pixbuf
   rs = gdk_pixbuf_get_rowstride(pixbuf);
   ppix = gdk_pixbuf_get_pixels(pixbuf);
   
   for (py = 0; py < hh; py++)                                             //  copy text pixels to pixbuf
   for (px = 0; px < ww; px++)
   {
      pix1 = cpix + 4 * (ww * py + px);
      pix2 = ppix + py * rs + 3 * px;
      if (bw == 'b')
         pix2[0] = pix2[1] = pix2[2] = 255 - pix1[3];                      //  black on white background
      else 
         pix2[0] = pix2[1] = pix2[2] = pix1[3];                            //  opposite
   }

   pango_font_description_free(pfont);
   g_object_unref(playout);
   cairo_destroy(cr);
   cairo_surface_destroy(surface);

   return pixbuf;
}


//  copy pixbuf1 into pixbuf2, treating color "rgb" as transparent
//  rgb = 0 = black, rgb = 255 = white
//  ftext = true to blend anti-aliased text edges (looks better)

void pixbuf_copy_alpha(GdkPixbuf *pxb1, int x1, int y1, int ww, int hh,
                       GdkPixbuf *pxb2, int x2, int y2, int rgb, int ftext)
{
   int         px, py, px1, py1, px2, py2;
   int         rs1, rs2;
   pixel       ppix1, ppix2, pix1, pix2;
   int         red1, green1, blue1, red2, green2, blue2;
   double      f1, f2;
   
   rs1 = gdk_pixbuf_get_rowstride(pxb1);
   ppix1 = gdk_pixbuf_get_pixels(pxb1);

   rs2 = gdk_pixbuf_get_rowstride(pxb2);
   ppix2 = gdk_pixbuf_get_pixels(pxb2);
   
   for (px = 0; px < ww; px++)
   for (py = 0; py < hh; py++)
   {
      px1 = x1 + px;                                                       //  copy-from pixel
      py1 = y1 + py;
      pix1 = ppix1 + rs1 * py1 + 3 * px1;

      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];

      if (red1 == rgb && green1 == rgb && blue1 == rgb)                    //  transparent, do not copy
         continue;
      
      px2 = x2 + px;                                                       //  copy-to pixel
      py2 = y2 + py;
      pix2 = ppix2 + rs2 * py2 + 3 * px2;

      if (ftext)
      {
         red2 = pix2[0];
         green2 = pix2[1];
         blue2 = pix2[2];
         
         f1 = abs(rgb - red1) / 255.0;                                     //  blend pixels (text)
         f2 = 1.0 - f1;
         red2 = int(f1 * red1 + f2 * red2);

         f1 = abs(rgb - green1) / 255.0;
         f2 = 1.0 - f1;
         green2 = int(f1 * green1 + f2 * green2);

         f1 = abs(rgb - blue1) / 255.0;
         f2 = 1.0 - f1;
         blue2 = int(f1 * blue1 + f2 * blue2);
         
         pix2[0] = red2;
         pix2[1] = green2;
         pix2[2] = blue2;
      }

      else                                                                 //  no blending (picture)  v.2.0
      {
         pix2[0] = red1;
         pix2[1] = green1;
         pix2[2] = blue1;
      }
   }

   return;
}


//  add a frame around a pixbuf image                                      //  new v.2.0
//  new pixbuf with added frame is returned
//  rgb: frame color   fww: thickness in pixels

GdkPixbuf * pixbuf_add_frame(GdkPixbuf *pixbuf1, int *rgb, int fww)
{
   GdkPixbuf   *pixbuf2;
   int         px, py;
   int         ww1, hh1, ww2, hh2, rs2;
   pixel       ppix2, pix2;
   
   ww1 = gdk_pixbuf_get_width(pixbuf1);                                    //  input pixbuf dimensions
   hh1 = gdk_pixbuf_get_height(pixbuf1);

   ww2 = ww1 + 2 * fww;                                                    //  new pixbuf with space for frame
   hh2 = hh1 + 2 * fww;
   pixbuf2 = gdk_pixbuf_new(colorspace,0,8,ww2,hh2);
   if (! pixbuf2) zappcrash("cannot create pixbuf");
   rs2 = gdk_pixbuf_get_rowstride(pixbuf2);
   ppix2 = gdk_pixbuf_get_pixels(pixbuf2);
   
   gdk_pixbuf_copy_area(pixbuf1,0,0,ww1,hh1,pixbuf2,fww,fww);              //  copy input pixbuf to output
   
   for (py = 0; py < fww; py++)                                            //  color frame, top
   for (px = 0; px < ww2; px++)
   {
      pix2 = ppix2 + rs2 * py + 3 * px;
      pix2[0] = rgb[0];
      pix2[1] = rgb[1];
      pix2[2] = rgb[2];
   }
   
   for (py = 0; py < hh2; py++)                                            //  right
   for (px = ww2 - fww; px < ww2; px++)
   {
      pix2 = ppix2 + rs2 * py + 3 * px;
      pix2[0] = rgb[0];
      pix2[1] = rgb[1];
      pix2[2] = rgb[2];
   }
   
   for (py = hh2 - fww; py < hh2; py++)                                    //  bottom
   for (px = 0; px < ww2; px++)
   {
      pix2 = ppix2 + rs2 * py + 3 * px;
      pix2[0] = rgb[0];
      pix2[1] = rgb[1];
      pix2[2] = rgb[2];
   }
   
   for (py = 0; py < hh2; py++)                                            //  left
   for (px = 0; px < fww; px++)
   {
      pix2 = ppix2 + rs2 * py + 3 * px;
      pix2[0] = rgb[0];
      pix2[1] = rgb[1];
      pix2[2] = rgb[2];
   }
   
   return pixbuf2;
}



